// (C) 2003-2015 Doctor Web. All rights reserved.

#include "drweb_ipc.h"

ipc_t::ipc_t(boost::asio::io_service& io) : io_(io)
{
}

ipc_t::~ipc_t()
{
}

void ipc_t::async_connect(const std::string& path)
{
    assert(!socket_);

    using namespace std::placeholders;
    socket_.reset(new boost::asio::local::stream_protocol::socket(io_));
    const boost::asio::local::stream_protocol::endpoint endpoint(path);
    socket_->async_connect(endpoint,
                           std::bind(&ipc_t::handle_async_connect, this_as_ipc(), _1));
}

void ipc_t::async_send_message(const google::protobuf::Message& msg)
{
    assert(msg.IsInitialized());

    static const uint32_t header_size = sizeof(uint32_t);
    const uint32_t message_size = msg.ByteSize();

    using namespace google::protobuf::io;
    send_buffer_.resize(send_buffer_.size() + header_size);
    CodedOutputStream::WriteLittleEndian32ToArray(
        message_size, &send_buffer_[send_buffer_.size() - header_size]);

    send_buffer_.resize(send_buffer_.size() + message_size);
    msg.SerializeWithCachedSizesToArray(&send_buffer_[send_buffer_.size() - message_size]);

    if(socket_ && socket_->is_open()) write_next();
}

void ipc_t::close_ipc()
{
    if(socket_) socket_->close();
}

void ipc_t::handle_async_connect(const boost::system::error_code& ec)
{
    if(ec) {
        if(ec == boost::asio::error::operation_aborted) return;
        throw boost::system::system_error(ec, "async_connect");
    }

    read_message_size();
    write_next();
}

void ipc_t::read_message_size()
{
    if(!is_open()) return;

    using namespace std::placeholders;
    receive_buffer_.resize(sizeof(uint32_t));
    boost::asio::async_read(*socket_, boost::asio::buffer(receive_buffer_),
                            std::bind(&ipc_t::handle_read_message_size, this_as_ipc(), _1,
                                      _2));
}

void ipc_t::handle_read_message_size(const boost::system::error_code& ec, size_t size)
{
    if(ec) {
        if(ec == boost::asio::error::operation_aborted) return;
        throw boost::system::system_error(ec, "read_message_size");
    }

    assert(size == sizeof(uint32_t));
    assert(size == receive_buffer_.size());
    using namespace google::protobuf::io;
    uint32_t message_size;
    CodedInputStream::ReadLittleEndian32FromArray(&receive_buffer_[0], &message_size);
    read_message(message_size);
}

void ipc_t::read_message(size_t message_size)
{
    if(!message_size) throw std::domain_error("bad message size");

    using namespace std::placeholders;
    receive_buffer_.resize(message_size);
    boost::asio::async_read(*socket_, boost::asio::buffer(receive_buffer_),
                            std::bind(&ipc_t::handle_read_message, this_as_ipc(), _1, _2));
}

void ipc_t::handle_read_message(const boost::system::error_code& ec, size_t size)
{
    if(ec) {
        if(ec == boost::asio::error::operation_aborted) return;
        throw boost::system::system_error(ec, "read_message");
    }

    assert(size == receive_buffer_.size());
    using namespace google::protobuf::io;
    ArrayInputStream is(&receive_buffer_[0], receive_buffer_.size());
    message_received(&is);

    read_message_size();
}

void ipc_t::write_next()
{
    if(!is_open()) return;

    if(sending_buffer_.empty() && !send_buffer_.empty()) {
        using namespace std::placeholders;

        sending_buffer_.swap(send_buffer_);
        boost::asio::async_write(*socket_, boost::asio::buffer(sending_buffer_),
                                 std::bind(&ipc_t::handle_write, this_as_ipc(), _1, _2));
    }
}

void ipc_t::handle_write(const boost::system::error_code& ec, size_t size)
{
    if(ec) {
        if(ec == boost::asio::error::operation_aborted) return;
        throw boost::system::system_error(ec, "write");
    }

    assert(size == sending_buffer_.size());
    sending_buffer_.clear();
    write_next();
}

bool ipc_t::is_open() const
{
    return socket_ && socket_->is_open();
}
