#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Example of synchronous DrWeb NetCheck client
"""

import os
import socket
import struct
import sys

import drweb_netcheck_api_pb2 as netcheck

CONFIGD_SOCKET = '/var/run/.com.drweb.public'
SIZE_STRUCT = '<I'
files = [ '/bin/sh', '/bin/ls', '/bin/rm' ]

def get_ncheck_socket():
    if len(sys.argv) != 2:
        print('Usage: {} <socket path>'.format(sys.argv[0]))
        exit(1)
    return sys.argv[1]

class netcheck_ipc(object):
    def __init__(self, socket_path):
        self.serial = 0
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(socket_path)
        self.send_hello()

    def send_hello(self):
        message = netcheck.ClientMessage()
        message.tag = message.HELLO
        message.origin = netcheck.APP_GENERIC
        message.origin_pid = os.getpid()
        self.send_message(message)

    def scan(self, path):
        job_id = self.send_init()
        self.send_file_by_chunks(job_id, path)
        self.send_scan(job_id)
        response = self.recv_scan_response(job_id)
        print(response)
        if response.cured and not response.cured_by_deletion:
            self.recv_cured_data(job_id)
        self.send_reset(job_id)

    def send_init(self):
        job_id = self.serial
        self.serial += 1
        message = netcheck.ClientMessage()
        message.tag = message.INIT
        message.job_id = job_id
        self.send_message(message)
        return job_id

    def send_file_by_chunks(self, job_id, path):
        fin = open(path, 'rb')
        while True:
            chunk = fin.read(4096)
            if not chunk:
                break
            self.send_data_chunk(job_id, chunk)
        fin.close()

    def send_data_chunk(self, job_id, chunk):
        message = netcheck.ClientMessage()
        message.tag = message.DATA_CHUNK
        message.job_id = job_id
        message.data = chunk
        self.send_message(message)

    def send_scan(self, job_id):
        message = netcheck.ClientMessage()
        message.tag = message.SCAN_REQUEST
        message.job_id = job_id
        message.scan_options.CopyFrom(netcheck.ScanOptions())
        message.scan_options.cure = True
        self.send_message(message)

    def send_get_data_request(self, job_id):
        message = netcheck.ClientMessage()
        message.tag = message.GET_DATA_REQUEST
        message.job_id = job_id
        self.send_message(message)

    def send_reset(self, job_id):
        message = netcheck.ClientMessage()
        message.tag = message.RESET
        message.job_id = job_id
        self.send_message(message)

    def recv_scan_response(self, job_id):
        response = self.recv_message()
        assert response.tag == netcheck.ServerMessage().SCAN_RESPONSE
        assert response.HasField('scan_report')
        assert response.job_id == job_id
        return response.scan_report

    def recv_cured_data(self, job_id):
        self.send_get_data_request(job_id)
        size = self.recv_get_data_response(job_id)
        cured_data = ""
        while len(cured_data) < size:
            buffer = self.recv_data_chunk(job_id)
            cured_data += buffer
            print("Cured data received job_id: {0}, size: {1}, remains: {2}".\
                format(job_id, len(buffer), size - len(cured_data)))

    def recv_get_data_response(self, job_id):
        response = self.recv_message()
        assert response.tag == netcheck.ServerMessage().GET_DATA_RESPONSE
        assert response.HasField('data_size')
        assert response.job_id == job_id
        return response.data_size

    def recv_data_chunk(self, job_id):
        data_chunk = self.recv_message()
        assert data_chunk.tag == netcheck.ServerMessage().DATA_CHUNK
        assert data_chunk.HasField('data')
        assert data_chunk.job_id == job_id
        return data_chunk.data

    def recv_message(self):
        size = struct.unpack(SIZE_STRUCT,
                             self.sock.recv(struct.calcsize(SIZE_STRUCT)))
        size = size[0]
        buffer = b''
        while len(buffer) < size:
            buffer += self.sock.recv(size - len(buffer))

        message = netcheck.ServerMessage()
        message.ParseFromString(buffer)
        return message

    def send_message(self, message):
        assert message.IsInitialized()
        buffer = struct.pack(SIZE_STRUCT, message.ByteSize())
        buffer += message.SerializeToString()
        while len(buffer):
            send = self.sock.send(buffer)
            buffer = buffer[send:]

ipc = netcheck_ipc(get_ncheck_socket())
for f in files:
    ipc.scan(f)
