// (C) 2003-2015 Doctor Web. All rights reserved.

#include "drweb_netcheck_scanner.h"

using namespace dw;
using namespace nc;

scanner_t::scanner_t(const std::shared_ptr<netcheck_api_t>& api) : api_(api)
{
}

void scanner_t::start(const std::vector<std::string>& files)
{
    files_.assign(files.begin(), files.end());
    send_init();
    if(jobs_.empty()) api_->stop();
}

void scanner_t::send_init()
{
    for(const auto& file : files_) {
        job_t job;
        job.file = file.c_str();
        job.fd = open(job.file, O_RDONLY);
        if(job.fd == -1) throw std::runtime_error("open() failed");
        job.cured_size = 0;
        jobs_.emplace(api_->init_job(this), job);
    }
}

void scanner_t::send_data()
{
    bool has_work;
    do {
        has_work = false;
        for(auto& value: jobs_) {
            const uint32_t job_id = value.first;
            job_t& job = value.second;
            if(job.fd == -1) continue;

            has_work = true;

            assert(api_->window_size() > 0);
            size_t window_size = api_->window_size();
            uint8_t buffer[4096];
            ssize_t part = read(job.fd, buffer, std::min(window_size, sizeof(buffer)));
            switch(part) {
                case -1:
                    throw std::runtime_error("read() failed");
                case 0: {
                    close(job.fd);
                    job.fd = -1;
                    ScanOptions options;
                    options.set_cure(true);
                    api_->send_scan_request(job_id, options);
                    break;
                }
                default:
                    api_->send_data(job_id, buffer, part);
                    break;
            }
            if(api_->window_size() <= 0) {
                has_work = false;
                break;
            }
        }
    } while(has_work);
}

// netcheck_api_t::delegate_t
void scanner_t::on_scan_report(uint32_t job_id, const ScanReport& report)
{
    auto found = jobs_.find(job_id);
    if(found != jobs_.end()) {
        const job_t& job = found->second;
        std::cout << "Scan report for file: " << job.file << std::endl;
        std::cout << report.DebugString() << std::endl;

        if(report.cured() && !report.cured_by_deletion())
            api_->send_get_data_request(job_id);
        else {
            api_->send_reset(job_id);
            jobs_.erase(found);
            if(jobs_.empty()) api_->stop();
        }
    }
}

void scanner_t::on_cured_data_size(uint32_t job_id, off_t size)
{
    auto found = jobs_.find(job_id);
    if(found != jobs_.end()) {
        job_t& job = found->second;
        std::cout << "Cured total data size for file: " << job.file << ": " << size
                  << std::endl;
        job.cured_size = size;
    }
}

void scanner_t::on_cured_data(uint32_t job_id, const void* data, size_t size)
{
    auto found = jobs_.find(job_id);
    if(found != jobs_.end()) {
        job_t& job = found->second;
        assert(job.cured_size >= static_cast<off_t>(size));
        job.cured_size -= size;
        std::cout << "Cured chunk data size for file: " << job.file << ": "
                  << " remains: " << job.cured_size << std::endl;

        // All data received
        if(!job.cured_size) {
            api_->send_reset(job_id);
            jobs_.erase(found);
            if(jobs_.empty()) api_->stop();
        }
    }
}

void scanner_t::on_alarm(uint32_t job_id, ErrorCode ec)
{
    auto found = jobs_.find(job_id);
    if(found != jobs_.end()) {
        const job_t& job = found->second;
        std::cout << "Alarm for file: " << job.file << ": " << ec << std::endl;
        jobs_.erase(found);
        if(jobs_.empty()) api_->stop();
    }
}

void scanner_t::on_window_replenished()
{
    if(api_->window_size() <= 0) return;
    send_data();
}
